from logging import WARN, DEBUG, INFO
from enum import Enum

import ldap3
from ldap3 import Connection, Server, ServerPool
from sentry_sdk import add_breadcrumb

from ..log import make_logger


class AuthLDAPResult(Enum):
    CONNECTION_FAILED_UNKNOWN = 1
    AUTH_FAILED = 2
    SUCCESS = 3
    BIND_USER_AUTH_FAILED = 4


class AuthLDAP(object):
    def __init__(self, app, config_obj=None, init_config=True):
        self.app = app
        self.log = make_logger(
            __name__, level=DEBUG if self.app.debug else INFO)

        def log_exception(txt, **kwargs):
            return self.log.error(txt, exc_info=True, **kwargs)
        self.handle_exception = log_exception

        if init_config:
            self.init_config(config_obj=config_obj)

    def init_config(self, config_obj=None):
        if not config_obj:
            config_obj = self.app.config
        self.log.info(f"Initialising LDAP settings from <{config_obj.__module__}.{type(config_obj).__name__}>")
        self.use_ldap = config_obj.get('ENABLE_LDAP')
        if self.use_ldap:
            ldap_host = config_obj.get('LDAP_HOST')
            ldap_port = config_obj.get('LDAP_PORT')
            self.base_dn = config_obj.get('LDAP_BASE_DN')
            self.id_attribute = config_obj.get('LDAP_USER_LOGIN_ATTR')

            if ldap_host and ldap_port:
                self._server = Server('{}:{}'.format(ldap_host, ldap_port),
                                      get_info=ldap3.ALL)

            self.log.info(f"Using LDAP server: {ldap_host}:{ldap_port}")
        else:
            self.log.info('LDAP is disabled.')

    def get_user_info(self, username, connection):
        # Using a bound connection, we will go fetch some LDAP attributes
        # for the user that logged in.
        connection.search(
            search_base=self.base_dn,
            search_filter='({}={})'.format(
                self.id_attribute, username),
            search_scope=ldap3.SUBTREE,
            attributes=[ldap3.ALL_ATTRIBUTES, ldap3.ALL_OPERATIONAL_ATTRIBUTES]
        )

        add_breadcrumb(
            category='ldap_auth',
            data=connection.response,
            level='debug',
        )

        filtered_response = [x for x in connection.response if 'dn' in x]

        if len(filtered_response) < 1:
            # Found no results
            self.log.error(f"Received invalid LDAP response, expected 1 got none: {connection.response} for username {username}")
            return None
        elif len(filtered_response) > 1:
            self.log.error(f"Received invalid LDAP response, expected 1 got {len(filtered_response)}: {filtered_response} for username {username}")
            return None

        return connection.response[0]

    def authenticate(self, username, password):
        if not self.use_ldap:
            return AuthLDAPResult.AUTH_FAILED, None

        connection = Connection(
            self._server,
            user=f'detnsw\\{username}',
            password=password,
            authentication=ldap3.NTLM,
            raise_exceptions=True,
            read_only=True,
            client_strategy=ldap3.RESTARTABLE
        )

        num_retries = self.app.config.get('LDAP_NUM_RETRY', 3)
        for _ in range(0, num_retries):
            try:
                connection.bind()
                user_data = self.get_user_info(username, connection)
                self.log.debug(
                    f"Authentication succeeded for user {repr(username)}")
            except ldap3.core.exceptions.LDAPInvalidCredentialsResult:
                self.log.info(
                    f"Authentication failed for user {repr(username)}")
                return AuthLDAPResult.AUTH_FAILED, None
            except ldap3.core.exceptions.LDAPCommunicationError as e:
                self.handle_exception(
                    f"Authentication failed due to LDAPCommunicationError: {connection.last_error}")
            except (ldap3.core.exceptions.LDAPServerPoolExhaustedError,
                    ldap3.core.exceptions.LDAPMaximumRetriesError) as e:
                self.handle_exception(
                    f"Authentication failed due to bad connection to LDAP server: {e}")
            else:
                # Successfully authenticated.
                if user_data:
                    return AuthLDAPResult.SUCCESS, user_data

                self.log.exception(
                    f"We should not have gotten here. user_data is invalid ({user_data}), check your search dn?")
                break
            finally:
                try:
                    connection.sasl_in_progress = False
                    connection.unbind()
                except:  # noqa
                    self.handle_exception("Unknown exception.")

        self.log.exception(
            f"We should not have gotten here either. Failed after {num_retries} retries.")
        return AuthLDAPResult.CONNECTION_FAILED_UNKNOWN, None

    @property
    def connection_established(self):
        if not self.use_ldap:
            return False

        connection = Connection(self._server, raise_exceptions=True, read_only=True)
        try:
            connection.bind()
        except ldap3.core.exceptions.LDAPSocketOpenError:
            return False
        except ldap3.core.exceptions.LDAPServerPoolExhaustedError:
            return False
        except ldap3.core.exceptions.LDAPOperationsErrorResult:
            return True
        except ldap3.core.exceptions.LDAPInvalidCredentialsResult:
            return True

        return True



import json


class ServerSentEvent(object):

    def __init__(self, data):
        self.data = data
        if isinstance(self.data, (dict, list)):
            self.data = json.dumps(self.data)
        self.event = None
        self.id = None
        self.desc_map = {
            'data': self.data,
            'event': self.event,
            'id': self.id
        }

    def encode(self):
        if not self.data:
            return ''

        lines = ["%s: %s" % (k, v)
                 for k, v in self.desc_map.items() if v]

        return "%s\n\n" % "\n".join(lines)



import json
from typing import Union

from sqlalchemy.types import TypeDecorator, Text


JSONEncodable = Union[dict, list, str]


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string."""

    impl = Text

    def process_bind_param(self, value: JSONEncodable, dialect) -> str:
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value: str, dialect) -> JSONEncodable:
        if value is not None:
            value = json.loads(value)
        return value


def base36encode(number, alphabet='0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'):
    """Converts an integer to a base36 string."""
    if not isinstance(number, int):
        raise TypeError('number must be an integer')

    base36 = ''
    sign = ''

    if number < 0:
        sign = '-'
        number = -number

    if 0 <= number < len(alphabet):
        return sign + alphabet[number]

    while number != 0:
        number, i = divmod(number, len(alphabet))
        base36 = alphabet[i] + base36

    return sign + base36


def base36decode(number):
    return int(number, 36)


from datetime import datetime
from sqlalchemy import Column, Integer, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from flask_login import current_user


class AuditMixin(object):
    created_at = Column(DateTime, default=datetime.now)
    updated_at = Column(DateTime, default=datetime.now, onupdate=datetime.now)

    @declared_attr
    def created_by_id(cls):
        return Column(Integer,
            ForeignKey('user.id', name='fk_%s_created_by_id' % cls.__name__, use_alter=True),
            # nullable=False,
            default=_current_user_id_or_none
        )

    @declared_attr
    def created_by(cls):
        return relationship(
            'User',
            primaryjoin='User.id == %s.created_by_id' % cls.__name__,
            remote_side='User.id'
        )

    @declared_attr
    def updated_by_id(cls):
        return Column(Integer,
            ForeignKey('user.id', name='fk_%s_updated_by_id' % cls.__name__, use_alter=True),
            # nullable=False,
            default=_current_user_id_or_none,
            onupdate=_current_user_id_or_none
        )

    @declared_attr
    def updated_by(cls):
        return relationship(
            'User',
            primaryjoin='User.id == %s.updated_by_id' % cls.__name__,
            remote_side='User.id'
        )


def _current_user_id_or_none():
    try:
        return current_user.id
    except:
        return None



def patch_uuid():
    import uuid

    global StringCompareUUID
    class StringCompareUUID(uuid.UUID):
        r"""Monkeypatched UUID class so we can check them directly
        against strings for convenience, shouldn't have any weird
        side-effects."""

        def __eq__(self, obj):
            if isinstance(obj, (str, bytes,)):
                try:
                    obj = uuid.UUID(obj)
                except ValueError:
                    pass
            return super().__eq__(obj)
        
        def __hash__(self):
            return hash(self.int)

    StringCompareUUID.__name__ = uuid.UUID.__name__
    uuid.UUID = StringCompareUUID



import os
import importlib

from .ReverseProxied import ReverseProxied


def load_config_cls(config_module, *args, **kwargs):
    env_config_classname = os.environ.get('CONFIG_CLASS', 'Development') + 'Config'

    config_cls = getattr(importlib.import_module(config_module), env_config_classname)
    config_obj = config_cls()

    return config_obj

__all__ = ['ReverseProxied', 'load_config_cls']



import logging
from typing import Optional
from functools import wraps

import msgpack
from sqlalchemy.exc import SQLAlchemyError, ProgrammingError

from ..log import make_logger

try:
    from flask import has_app_context
except ImportError:
    def has_app_context():
        return False


def teardown_db(f):
    @wraps(f)
    def wrapper(self, *args, **kwds):
        try:
            result = f(self, *args, **kwds)
        except SQLAlchemyError as e:
            self.session.rollback()
            raise e
        if not has_app_context():
            self.log.debug('Closing session manually as we\'re outside of app context.')
            self.session.remove()
        return result
    return wrapper


class UserConfigManager(object):
    def __init__(self, session, kv_model, defaults={}, strict=True):
        self.kv_model = kv_model
        self.session = session
        self.defaults = defaults
        self.strict = strict
        self.cache = {}

        self.log = make_logger(__name__)
    
    @teardown_db
    def get(self, key):
        key = key.lower()
        if key in self.cache:
            self.log.debug(f"Loaded db_config key from cache: {key}={self.cache[key]}")
            return self.cache[key]

        kv: Optional[self.kv_model] = None
        try:
            kv = self.session.query(self.kv_model).get(key)
        except ProgrammingError:
            self.session.rollback()

        if kv:
            # Key was found in db, good to go.
            cv = self.decode(kv.value)
            self.cache[key] = cv
            return cv

        # Key isn't in db, check defaults.
        if key in self.defaults:
            self.log.debug(f"Loaded db_config key from defaults: {key}={self.defaults[key]}")
            # Got a default, return it.
            return self.defaults[key]
        elif self.strict:
            # Item wasn't previously set and doesn't have a default.
            raise KeyError
        else:
            self.log.error(f"Tried to load missing db_config key: {key}")
            return None

    @teardown_db
    def set(self, key, value):
        key = key.lower()
        kv = self.session.query(self.kv_model).get(key)
        if not kv:
            kv = self.kv_model(key=key)
            self.session.add(kv)

        kv.value = self.encode(value)
        self.session.commit()

        self.log.debug(f'Wrote db_config key to database: {key}={value}')
        self.cache[kv.key] = value

        return kv.value

    # Return a dict of all settings, including defaults.
    @teardown_db
    def get_all(self):
        self.cache = {
            r.key: self.decode(r.value)
            for r in self.kv_model.query.all()
        }
        return {**self.defaults, **self.cache}

    # Remove the key from the db, effectively resetting to the default value.
    @teardown_db
    def reset(self, key):
        kv = self.kv_model.query.get(key)
        if kv:
            self.session.delete(kv)
            self.session.commit()
        self.cache.pop(key, None)
        self.log.debug(f'Reset db_config key ({key}).')

    # Encode the value before inserting into the db.
    def encode(self, value):
        return msgpack.packb(value, use_bin_type=True)

    # Decode the value from the db.
    def decode(self, value):
        return msgpack.unpackb(value, raw=False)



from typing import Iterator, Tuple, Optional
import re


class FilterParser(object):
    expression = re.compile(r'(?:([\w\d]+):(?:"(.+)"|((?!")\S+)))')

    def match(self, key: str, query_str: str) -> Optional[Tuple[str, str]]:
        for m_key, m_value in self.match_all(query_str):
            if m_key == key:
                return m_value
        return None

    def match_all(self, query_str: str) -> Iterator[Tuple[str, str]]:
        for res in self.expression.finditer(query_str):
            val = res.group(2) or res.group(3)
            yield res.group(1), val

    def clear_filters(self, query_str: str) -> str:
        return self.expression.sub('', query_str).strip()



import json
import os
import zipfile
from io import BytesIO
from tempfile import TemporaryDirectory
from hashlib import sha512
import time

from PIL import Image
from OpenSSL import crypto

from ..log import make_logger


class PushPackageBuilder(object):

    def __init__(self, icon_path, website_dict, key_path, key_password=None):
        self.images = {}
        self.website_dict = website_dict
        self.build_images(icon_path)
        with open(key_path, 'rb') as f:
            key_buf = f.read()
        self.cert = crypto.load_certificate(crypto.FILETYPE_PEM, key_buf)
        self.key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_buf)
        self.log = make_logger(__name__)

    def get_hash_details(self, data=None, sha_hash=None):
        return {
            'hashType': 'sha512',
            'hashValue': sha_hash or sha512(data).hexdigest()
        }

    def build_images(self, image_src, force=False):
        sizecats = (16, 32, 128)
        if len(self.images.values()) == sizecats * 2 and not force:
            # Use cached images.
            return
        else:
            # Reset self.images, we want to regen.
            self.images = {}

        with Image.open(image_src) as im:
            for sizecat in sizecats:
                with BytesIO() as nim_f:
                    nim = im.resize((sizecat, sizecat))
                    nim.save(nim_f, format='png')
                    self.images[f'icon_{sizecat}x{sizecat}.png'] = nim_f.getvalue()

                with BytesIO() as nim_f:
                    nim = im.resize((sizecat*2, sizecat*2))
                    nim.save(nim_f, format='png')
                    self.images[f'icon_{sizecat}x{sizecat}@2x.png'] = nim_f.getvalue()

    def build_zip(self, zf, website_dict):
        manifest = {}
        for path, data in self.images.items():
            full_path = os.path.join('icon.iconset/', path)
            zf.writestr(full_path, data)
            manifest[full_path] = self.get_hash_details(data)

        website_data = json.dumps(website_dict).encode()
        zf.writestr('website.json', website_data)
        manifest['website.json'] = self.get_hash_details(website_data)

        manifest_data = json.dumps(manifest).encode()
        zf.writestr('manifest.json', manifest_data)
        self.log.debug('Wrote manifest: %s', manifest_data)

        zf.writestr('signature', self.sign_manifest(manifest_data))

    def build_pushpackage(self, output_file=None, merge_website_dict={}):
        output_file = output_file or BytesIO()

        with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zf:
            self.log.debug('Building new zip in %s', zf)
            self.build_zip(zf, website_dict={
                **self.website_dict,
                **merge_website_dict
            })
        output_file.seek(0)

        return output_file

    def sign_manifest(self, manifest_data):
        # Fuck me. Thanks @WhyNotHugo: https://stackoverflow.com/a/41553623
        bio_in = crypto._new_mem_buf(manifest_data)
        PKCS7_NOSIGS = 0x4  # defined in pkcs7.h
        pkcs7 = crypto._lib.PKCS7_sign(self.cert._x509, self.key._pkey, crypto._ffi.NULL, bio_in, PKCS7_NOSIGS)  # noqa
        bio_out = crypto._new_mem_buf()
        crypto._lib.i2d_PKCS7_bio(bio_out, pkcs7)
        sigbytes = crypto._bio_to_string(bio_out)

        return sigbytes



from dataclasses import dataclass
from math import floor, ceil
from colorsys import hls_to_rgb


@dataclass
class Rect:
    x: float
    y: float
    width: float
    height: float


class IdenticonGenerator(object):
    def __init__(self, background=(240, 240, 240, 1.0), margin=0.16,
                 image_size=64, saturation=0.8, lightness=0.5, inverted=False,
                 pixels=5):
        self.background = background
        self.margin = margin
        self.pixels = pixels
        self.image_size = image_size
        self.saturation = saturation
        self.lightness = lightness
        self.inverted = inverted
    
    def get_hash_color(self, hash):
        hue = int(hash[-7:], 16) / 0xfffffff
        return [i*255 for i in hls_to_rgb(hue, self.lightness, self.saturation)]
    
    def get_background(self, hash):
        if self.inverted:
            return self.get_hash_color(hash)
        return self.background
    
    def get_foreground(self, hash):
        if self.inverted:
            return self.background
        return self.get_hash_color(hash)
    
    def get_rectangles(self, hash):
        cols = ceil(self.pixels / 2)

        rectangles = []
        baseMargin = floor(self.image_size * self.margin)
        cell = floor((self.image_size - (baseMargin * 2)) / self.pixels)
        margin = floor((self.image_size - cell * self.pixels) / 2)
        background_str = self.get_rgba_string(*self.get_background(hash))
        foreground_str = self.get_rgba_string(*self.get_foreground(hash))

        for i in range(self.pixels * cols):
            if int(hash[i], 16) % 2:
                # If value is even, skip, this is blank.
                continue
            
            if i < self.pixels and self.pixels % 2:
                # Center column
                rectangles.append(Rect(
                    x=(cols - 1) * cell + margin,
                    y=i * cell + margin,
                    width=cell,
                    height=cell
                ))
            else:
                col = self.pixels * floor(i / self.pixels)
                distance = cols - (col / self.pixels)

                y = (i - col) * cell + margin
                rectangles.append(Rect(
                    x=((distance - 1) * cell) + margin,
                    y=y,
                    width=cell,
                    height=cell
                ))
                rectangles.append(Rect(
                    x=((self.pixels - distance) * cell) + margin,
                    y=y,
                    width=cell,
                    height=cell
                ))
            
        return rectangles

    def dump(self, hash):
        stroke = self.image_size * 0.005

        background_str = self.get_rgba_string(*self.get_background(hash))
        foreground_str = self.get_rgba_string(*self.get_foreground(hash))

        xml = (f"<svg xmlns='http://www.w3.org/2000/svg'"
                   f" width='{self.image_size}' height='{self.image_size}'"
                   f" style='background-color: {background_str}'>"
               f"<g style='fill: {foreground_str}; stroke: {foreground_str}; stroke-width: {stroke};'>")
        for rect in self.get_rectangles(hash):
            xml += (f"<rect x='{rect.x}'"
                         f" y='{rect.y}'"
                         f" width='{rect.width}'"
                         f" height='{rect.height}'/>")
        xml += "</g></svg>"

        return xml
    
    @classmethod
    def get_rgba_string(cls, red, green, blue, alpha=1.0):
        return f"rgba({round(red)}, {round(green)}, {round(blue)}, {alpha})"



from contextlib import ContextDecorator
from unittest.mock import patch

from flask import Flask, current_app, _request_ctx_stack
from flask_login import UserMixin, current_user, login_user, logout_user


class UserTestContext(ContextDecorator):
    r"""A context manager that executes a block of code within
    the context of a specific logged in user.
    Can also be used as a decorator."""

    def __init__(self, user: UserMixin, app: Flask = False):
        self.user = user
        self.app = app or current_app
        self.client = self.app.test_client()
    
    def __enter__(self):
        # Enter the flask test client context.
        client = self.client.__enter__()

        # Run the standard login method.
        with client.session_transaction() as session:
            # A hack so that login_user sees the correct session to modify.
            with patch('flask._request_ctx_stack.top.session', new=session):
                login_user(self.user, force=True, fresh=True)

        return client

    def __exit__(self, exc_type, exc_value, tb):
        try:
            # Logout of our user gracefully.
            with self.client.session_transaction() as session:
                with patch('flask._request_ctx_stack.top.session', new=session):
                    logout_user()
        finally:
            # Exit the flask test client context.
            #self.session_transaction.__exit__(exc_type, exc_value, tb)
            self.client.__exit__(exc_type, exc_value, tb)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fourutils',
 'fourutils.asynchronous',
 'fourutils.flask',
 'fourutils.flask.testing',
 'fourutils.ldap',
 'fourutils.sqla',
 'fourutils.sqla.mixins',
 'fourutils.sqla.types',
 'fourutils.webnotifications']

package_data = \
{'': ['*']}

install_requires = \
['sentry-sdk>=0,<1']

setup_kwargs = {
    'name': 'fourutils',
    'version': '0.2.1',
    'description': '',
    'long_description': None,
    'author': 'Oscar Rainford',
    'author_email': 'oscar@fourbs.com.au',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .log import make_logger
from .ServerSentEvent import ServerSentEvent
from .FilterParser import FilterParser
from .base36 import base36encode, base36decode
from .IdenticonGenerator import IdenticonGenerator


__all__ = ['ServerSentEvent', 'FilterParser', 'base36encode',
           'base36decode', 'IdenticonGenerator']
